//
//  AppDelegate.swift
//  BLDCMotor
//
//  Created by Phidgets on 2018-02-22.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = BLDCMotor()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var dataBox: NSBox!
    @IBOutlet weak var velocityLabel: NSTextField!
    @IBOutlet weak var positionLabel: NSTextField!
    @IBOutlet weak var brakingLabel: NSTextField!
    @IBOutlet weak var settingsBox: NSBox!
    @IBOutlet weak var rescaleText: NSTextField!
    @IBOutlet weak var targetVelocitySlider: NSSlider!
    @IBOutlet weak var minTargetVelocityLabel: NSTextField!
    @IBOutlet weak var maxTargetVelocityLabel: NSTextField!
    @IBOutlet weak var targetVelocityLabel: NSTextField!
    @IBOutlet weak var accelerationSlider: NSSlider!
    @IBOutlet weak var minAccelerationLabel: NSTextField!
    @IBOutlet weak var maxAccelerationLabel: NSTextField!
    @IBOutlet weak var accelerationLabel: NSTextField!
    @IBOutlet weak var stallVelocitySlider: NSSlider!
    @IBOutlet weak var minStallVelocityLabel: NSTextField!
    @IBOutlet weak var maxStallVelocityLabel: NSTextField!
    @IBOutlet weak var stallVelocityLabel: NSTextField!
    @IBOutlet weak var brakingStrengthSlider: NSSlider!
    @IBOutlet weak var minBrakingStrengthLabel: NSTextField!
    @IBOutlet weak var maxBrakingStrengthLabel: NSTextField!
    @IBOutlet weak var brakingStrengthLabel: NSTextField!
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0

    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.velocityUpdate.addHandler(velocityupdate_handler)
            let _ = ch.positionChange.addHandler(positionchange_handler)
            let _ = ch.brakingStrengthChange.addHandler(brakingchange_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.positionChange.removeAllHandlers()
            ch.velocityUpdate.removeAllHandlers()
            ch.brakingStrengthChange.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! BLDCMotor
            let minTargetVelocity = try attachedDevice.getMaxVelocity() * -1
            let maxTargetVelocity = try attachedDevice.getMaxVelocity()
            let targetVelocity = try attachedDevice.getMinVelocity()
            let minAcceleration = try attachedDevice.getMinAcceleration()
            let maxAcceleration = try attachedDevice.getMaxAcceleration()
            let acceleration = try attachedDevice.getAcceleration()
            let minStallVelocity = try attachedDevice.getMinStallVelocity()
            let maxStallVelocity = try attachedDevice.getMaxStallVelocity()
            let stallVelocity = try attachedDevice.getStallVelocity()
            let minBrakingStrength = try attachedDevice.getMinBrakingStrength()
            let maxBrakingStrength = try attachedDevice.getMaxBrakingStrength()
            let brakingStrength = try attachedDevice.getMinBrakingStrength()
            let rescale = try attachedDevice.getRescaleFactor()
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.settingsBox.isHidden = false
                self.dataBox.isHidden = false
                
                self.rescaleText.stringValue = String(format:"%.3f",rescale)
                self.minTargetVelocityLabel.stringValue = String(format:"%.3f", minTargetVelocity)
                self.maxTargetVelocityLabel.stringValue = String(format:"%.3f", maxTargetVelocity)
                self.targetVelocityLabel.stringValue = String(format:"%.3f", targetVelocity)
                self.targetVelocitySlider.minValue = minTargetVelocity
                self.targetVelocitySlider.maxValue = maxTargetVelocity
                self.targetVelocitySlider.doubleValue = targetVelocity
                
                self.minAccelerationLabel.stringValue = String(format:"%.3f", minAcceleration)
                self.maxAccelerationLabel.stringValue = String(format:"%.3f", maxAcceleration)
                self.accelerationLabel.stringValue = String(format:"%.3f", acceleration)
                self.accelerationSlider.minValue = minAcceleration
                self.accelerationSlider.maxValue = maxAcceleration
                self.accelerationSlider.doubleValue = acceleration
                
                self.stallVelocitySlider.minValue = minStallVelocity
                self.stallVelocitySlider.maxValue = maxStallVelocity
                self.stallVelocitySlider.doubleValue = stallVelocity
                self.minStallVelocityLabel.stringValue = String(format:"%.3f", minStallVelocity)
                self.maxStallVelocityLabel.stringValue = String(format:"%.3f", maxStallVelocity)
                self.stallVelocityLabel.stringValue = String(format:"%.3f", stallVelocity)
                
                self.minBrakingStrengthLabel.stringValue = String(format:"%.3f", minBrakingStrength)
                self.maxBrakingStrengthLabel.stringValue = String(format:"%.3f", maxBrakingStrength)
                self.brakingStrengthLabel.stringValue = String(format:"%.3f", brakingStrength)
                self.brakingStrengthSlider.minValue = minBrakingStrength
                self.brakingStrengthSlider.maxValue = maxBrakingStrength
                self.brakingStrengthSlider.doubleValue = brakingStrength
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.settingsBox.isHidden = true
            self.dataBox.isHidden = true
            
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func positionchange_handler(sender: BLDCMotor, position: Double){
        DispatchQueue.main.async {
            self.positionLabel.stringValue = String(format:"%.3f", position)
        }
    }
    
    func velocityupdate_handler(sender: BLDCMotor, velocity: Double){
        DispatchQueue.main.async {
            self.velocityLabel.stringValue = String(format:"%.3f", velocity)
        }
    }
    
    func brakingchange_handler(sender: BLDCMotor, brakingStrength: Double){
        DispatchQueue.main.async {
            self.brakingLabel.stringValue = String(format:"%.3f", brakingStrength)
        }
    }
    
    //MARK: GUI Controls
    @IBAction func setRescale(_ sender: Any) {
        do{
            if let rescale = Double(rescaleText.stringValue){
                try ch.setRescaleFactor(rescale)
                updateGUI()
            }
            else{
                rescaleText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func stop(_ sender: Any) {
        do{
            try ch.setTargetVelocity(0)
            targetVelocitySlider.doubleValue = 0.0
            targetVelocityLabel.stringValue = "0.000"
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func zeroPosition(_ sender: Any) {
        do{
            var position = try ch.getPosition()
            try ch.addPositionOffset(positionOffset: -1 * position)
            position = try ch.getPosition()
            positionLabel.stringValue = String(format:"%.3f", position)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setTargetVelocity(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setTargetVelocity(slider.doubleValue)
            targetVelocityLabel.stringValue = String(format:"%.3f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setAcceleration(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setAcceleration(slider.doubleValue)
            accelerationLabel.stringValue = String(format:"%.3f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setStallVelocity(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setStallVelocity(slider.doubleValue)
            stallVelocityLabel.stringValue = String(format:"%.3f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setBrakingStrength(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setTargetBrakingStrength(slider.doubleValue)
            brakingStrengthLabel.stringValue = String(format:"%.3f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Misc
    func updateGUI(){
        do{
            let minStallVelocity = try ch.getMinStallVelocity()
            let maxStallVelocity = try ch.getMaxStallVelocity()
            let stallVelocity = try ch.getStallVelocity()
            let position = try ch.getPosition()
            
            stallVelocitySlider.minValue = minStallVelocity
            stallVelocitySlider.maxValue = maxStallVelocity
            stallVelocitySlider.doubleValue = stallVelocity
            minStallVelocityLabel.stringValue = String(format:"%.3f", minStallVelocity)
            maxStallVelocityLabel.stringValue = String(format:"%.3f", maxStallVelocity)
            stallVelocityLabel.stringValue = String(format:"%.3f", stallVelocity)
            positionLabel.stringValue = String(format:"%.3f", position)
            
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
}

